import 'dart:async';
import 'dart:io';

import '../ftp_exceptions.dart';
import '../ftp_reply.dart';
import '../ftp_socket.dart';
import '../util/transfer_util.dart';
import 'file.dart';

typedef void FileProgress(
    double progressInPercent, int totalReceived, int fileSize);

class FileDownload {
  final FTPSocket _socket;

  /// File Download Command
  FileDownload(this._socket);

  Future<bool> downloadFile(
    String? sRemoteName,
    File fLocalFile, {
    FileProgress? onProgress,
  }) async {
    _socket.logger.log('Download $sRemoteName to ${fLocalFile.path}');
    //check for file existence and init totalData to receive
    int fileSize = 0;
    fileSize = await FTPFile(_socket).size(sRemoteName);
    if (fileSize == -1) {
      throw FTPException('Remote File $sRemoteName does not exist!');
    }

    // Transfer Type
    await _socket.setTransferType(_socket.transferType);

    // Enter passive mode
    FTPReply response = await _socket.openDataTransferChannel();

    //the response will be the file, witch will be loaded with another socket
    _socket.sendCommandWithoutWaitingResponse('RETR $sRemoteName');

    // Data Transfer Socket
    int iPort = TransferUtil.parsePort(response.message, _socket.supportIPV6);
    _socket.logger.log('Opening DataSocket to Port $iPort');
    final Socket dataSocket = await Socket.connect(_socket.host, iPort,
        timeout: Duration(seconds: _socket.timeout));
    // Test if second socket connection accepted or not
    response = await _socket.readResponse();
    //some server return two lines 125 and 226 for transfer finished
    bool isTransferCompleted = response.isSuccessCode();
    if (!isTransferCompleted && response.code != 125 && response.code != 150) {
      throw FTPException('Connection refused. ', response.message);
    }

    // Changed to listen mode instead so that it's possible to send information back on downloaded amount
    _socket.logger.log('Start downloading...');
    var sink = fLocalFile.openWrite(mode: FileMode.writeOnly);
    var received = 0;
    await dataSocket.listen((data) {
      sink.add(data);
      if (onProgress != null) {
        received += data.length;
        var percent = ((received / fileSize) * 100).toStringAsFixed(2);
        //in case that the file size is 0, then pass directly 100
        double percentVal = double.tryParse(percent) ?? 100;
        if (percentVal.isInfinite || percentVal.isNaN) percentVal = 100;
        onProgress(percentVal, received, fileSize);
      }
    }).asFuture();

    await dataSocket.close();
    await sink.flush();
    await sink.close();

    if (!isTransferCompleted) {
      //Test if All data are well transferred
      response = await _socket.readResponse();
      if (!response.isSuccessCode()) {
        throw FTPException('Transfer Error.', response.message);
      }
    }

    _socket.logger.log('File Downloaded!');
    return true;
  }
}
