import 'dart:async';
import 'dart:io';

import 'package:path/path.dart';

import '../ftp_exceptions.dart';
import '../ftp_reply.dart';
import '../ftp_socket.dart';
import '../util/transfer_util.dart';
import 'file_download.dart';

class FileUpload {
  final FTPSocket _socket;

  /// File Upload Command
  FileUpload(this._socket);

  /// Upload File [fFile] to the current directory with [remoteName] (using filename if not set)
  Future<bool> uploadFile(
    File fFile, {
    String remoteName = '',
    FileProgress? onProgress,
  }) async {
    _socket.logger.log('Upload File: ${fFile.path}');

    await _socket.setTransferType(_socket.transferType);

    // Enter passive mode
    FTPReply response = await _socket.openDataTransferChannel();

    // Store File
    String sFilename = remoteName;
    if (sFilename.isEmpty) {
      sFilename = basename(fFile.path);
    }

    // The response is the file to upload, witch will be managed by another socket
    _socket.sendCommandWithoutWaitingResponse('STOR $sFilename');

    // Data Transfer Socket
    int iPort = TransferUtil.parsePort(response.message, _socket.supportIPV6);
    _socket.logger.log('Opening DataSocket to Port $iPort');
    final Socket dataSocket = await Socket.connect(_socket.host, iPort);
    //Test if second socket connection accepted or not
    response = await _socket.readResponse();
    //some server return two lines 125 and 226 for transfer finished
    bool isTransferCompleted = response.isSuccessCode();
    if (!isTransferCompleted && response.code != 125 && response.code != 150) {
      throw FTPException('Connection refused. ', response.message);
    }

    _socket.logger.log('Start uploading...');

    var received = 0;
    int fileSize = await fFile.length();

    Stream<List<int>> readStream = fFile.openRead().transform(
      StreamTransformer.fromHandlers(
        handleData: (data, sink) {
          sink.add(data);
          if (onProgress != null) {
            received += data.length;
            var percent = ((received / fileSize) * 100).toStringAsFixed(2);
            //in case that the file size is 0, then pass directly 100
            double percentVal = double.tryParse(percent) ?? 100;
            if (percentVal.isInfinite || percentVal.isNaN) percentVal = 100;
            onProgress(percentVal, received, fileSize);
          }
        },
      ),
    );

    await dataSocket.addStream(readStream);
    await dataSocket.flush();
    await dataSocket.close();

    if (!isTransferCompleted) {
      // Test if All data are well transferred
      response = await _socket.readResponse();
      if (!response.isSuccessCode()) {
        throw FTPException('Transfer Error.', response.message);
      }
    }

    _socket.logger.log('File Uploaded!');
    return true;
  }
}
